@@ -5,8 +5,8 @@ api_v2.py
 The core SDK module
 """
 
-
 from datetime import datetime
+from dateutil.parser import parse as dateparse
 from enum import Enum
 from http import HTTPStatus
 import json
@@ -99,7 +99,7 @@ class InferredFieldType(Enum):
         "type": "ENTITY",
         "isList": True,
     }
-    document_tag = {"expression": "document_tag(File)", "type": "STRING"}
+    document_tag = {"expression": "document_tag(File)", "type": "ENTITY", "isList": True}
 
     @property
     def expr(self):
@@ -167,15 +167,22 @@ class ResourceType(str, Enum):
     files = "files"
 
 
+FMTS = ["%Y-%m-%d", "%Y-%m", "%Y"]
+
+
 @validate_arguments
 def parse_date(s: str) -> datetime:
-    try:
-        return datetime.strptime(s, "%Y-%m-%d")
-    except ValueError:
+    for i, fmt in enumerate(FMTS):
         try:
-            return datetime.strptime(s, "%Y-%m")
-        except ValueError:
-            return datetime.strptime(s, "%Y")
+            return datetime.strptime(s, fmt)
+        except ValueError as e:
+            # From https://stackoverflow.com/questions/5045210/how-to-remove-unconverted-data-from-a-python-datetime-object
+            if len(e.args) > 0 and e.args[0].startswith("unconverted data remains: "):
+                prefix = s[: -(len(e.args[0]) - 26)]
+                return datetime.strptime(prefix, fmt)
+
+    # As a last ditch effort, try using dateparse
+    return dateparse(s)
 
 
 class Impira:
@@ -271,6 +278,7 @@ class Impira:
 
         resp = self.session.patch(
             self._build_resource_url("fc", collection_id),
+            params={"assert_updated": False},
             json={"data": data},
         )
 
@@ -578,6 +586,12 @@ class Impira:
                 args["timeout"] = math.ceil(timeout - time_since)
                 logging.warning("Request timed out, but still have %gs left. Will try again..." % (args["timeout"]))
                 continue
+            elif resp.status_code == HTTPStatus.TOO_MANY_REQUESTS and time_since < timeout - 1:
+                args["timeout"] = math.ceil(timeout - 1 - time_since)
+                logging.warning(
+                    "Hit a rate limit, but still have %gs left. Will sleep for 1s and try again..." % (args["timeout"])
+                )
+                time.sleep(1)
             else:
                 raise APIError(resp)
 
