@@ -70,18 +70,42 @@ def build_parser(subparsers, parent_parser):
         action="store_true",
         help="Only label existing fields in the collection",
     )
+    parser.add_argument(
+        "--label-empty-values",
+        default=False,
+        action="store_true",
+        help="If labels are missing, assign them 'No value' labels",
+    )
     parser.add_argument(
         "--max-fields",
         default=-1,
         type=int,
         help="Only create up to this many fields",
     )
+    parser.add_argument(
+        "--first-file",
+        default=-0,
+        type=int,
+        help="Start with this file (indexed by 0). Useful for pagination.",
+    )
     parser.add_argument(
         "--max-files",
         default=-1,
         type=int,
         help="Only upload up to this many files",
     )
+    parser.add_argument(
+        "--batch-size",
+        default=50,
+        type=int,
+        help="Batch size for downloading text and labeling files",
+    )
+    parser.add_argument(
+        "--first-batch",
+        default=0,
+        type=int,
+        help="First batch to start (this is useful for resuming)",
+    )
     Impira.add_arguments(parser)
 
     parser.set_defaults(func=main)
@@ -99,11 +123,13 @@ def main(args):
     manifest = DocManifest.parse_file(manifest_file)
     M = schema_to_model(manifest.doc_schema)
 
+    all_docs = []
     for doc in manifest.docs:
         doc.fname = workdir / doc.fname
-        assert doc.url is not None or doc.fname.exists()
-        if doc.record is not None:
-            doc.record = M.parse_obj(doc.record)
+        if doc.url is not None or doc.fname.exists():
+            all_docs.append(doc)
+            if doc.record is not None:
+                doc.record = M.parse_obj(doc.record)
 
     tool = Impira(config=Impira.Config(**vars(args)))
     tool.run(
@@ -117,7 +143,12 @@ def main(args):
         skip_missing_files=args.skip_missing_files,
         add_files=args.add_files,
         skip_new_fields=args.skip_new_fields,
+        empty_labels=args.label_empty_values,
         collection_name=args.name,
         max_fields=args.max_fields,
+        first_file=args.first_file,
         max_files=args.max_files,
+        batch_size=args.batch_size,
+        first_batch=args.first_batch,
+        cache_dir=workdir / "cache",
     )
