@@ -3,12 +3,13 @@ from concurrent.futures import ThreadPoolExecutor
 from enum import Enum
 import json
 import os
-from pydantic import BaseModel, validate_arguments
+import pathlib
+from pydantic import BaseModel, validate_arguments, ValidationError
 from typing import Any, Dict, List, Optional, Tuple, Type, Union
 import time
 from uuid import uuid4
 
-from impira import Impira as ImpiraAPI, FieldType, InferredFieldType, parse_date, APIError
+from .. import Impira as ImpiraAPI, FieldType, InferredFieldType, parse_date, APIError
 
 from ..cmd.utils import environ_or_required
 from ..schema import schema_to_model
@@ -88,9 +89,9 @@ class ImpiraWord(BaseModel):
     confidence: float
     location: Location
     processed_word: str
-    rotated: bool
-    source: str
-    uid: str
+    rotated: Optional[bool]
+    source: Optional[str]
+    uid: Optional[str]
     word: str
 
 
@@ -115,10 +116,19 @@ class CheckboxSource(BaseModel):
     BBoxes: List[Location]
 
 
+class DocumentTagValue(BaseModel):
+    class L(BaseModel):
+        Value: Optional[str]
+        IsPrediction: bool = False
+
+    Label: L
+
+
 class ScalarLabel(BaseModel):
     class L(BaseModel):
         Source: Optional[Union[CheckboxSource, List[ImpiraWord]]]
         IsPrediction: bool = False
+        IsConfident: Optional[bool] = None
         Value: Optional[Any]
 
     class C(BaseModel):
@@ -135,6 +145,7 @@ class ScalarLabel(BaseModel):
 class RowLabel(BaseModel):
     class L(BaseModel):
         IsPrediction: bool
+        IsConfident: Optional[bool] = None
         Value: Dict[str, ScalarLabel]
 
     Label: L
@@ -192,10 +203,15 @@ class EntityMap(BaseModel):
 
 
 @validate_arguments
-def find_overlapping_words(value, words: List[ImpiraWord]):
+def find_overlapping_words(value, word_map: Dict[str, ImpiraWord], words: List[ImpiraWord]):
     if value.location is None:
         return []
 
+    if len(value.location.uids) > 0:
+        uid_words = [word_map[uid] for uid in value.location.uids if uid in word_map]
+        if len(uid_words) == len(value.location.uids):
+            return uid_words
+
     return [
         word
         for word in words
@@ -210,13 +226,18 @@ def generate_labels(
     file_name: str,
     record,
     words: List[ImpiraWord],
+    word_map: Dict[str, ImpiraWord],
     entity_map: EntityMap,
     model_versions: Dict[str, int],
+    empty_labels: bool = False,
 ) -> Dict[str, Any]:
     labels = {}
     for field_name, value in dict(record).items():
         if isinstance(value, List):
-            rows = [generate_labels(log, file_name, v, words, entity_map, model_versions) for v in value]
+            rows = [
+                generate_labels(log, file_name, v, words, word_map, entity_map, model_versions, empty_labels)
+                for v in value
+            ]
             row_labels = [
                 RowLabel(
                     Label=RowLabel.L(
@@ -244,12 +265,24 @@ def generate_labels(
                     },
                 ),
                 Context=ScalarLabel.C(Entities=[]),
+                IsPrediction=False,
                 ModelVersion=model_versions.get(field_name, 0),
             )
 
             labels[field_name] = scalar_label
-        elif value is not None:
-            w = find_overlapping_words(value, words)
+        elif isinstance(value, DocumentTagLabel):
+            scalar_label = ScalarLabel(
+                Label=ScalarLabel.L(
+                    Source=[],
+                    Value=[DocumentTagValue(Label={"Value": x, "IsPrediction": False}) for x in value.fmt()],
+                ),
+                Context=ScalarLabel.C(Entities=[]),
+                IsPrediction=False,
+                ModelVersion=model_versions.get(field_name, 0),
+            )
+            labels[field_name] = scalar_label
+        elif value is not None and value.value is not None:
+            w = find_overlapping_words(value, word_map, words)
             candidate_entities = [e for x in w for e in entity_map.find_entities([x], match_supersets=True)]
             target_type = label_name_to_inferred_field_type(type(value).__name__)
 
@@ -282,6 +315,13 @@ def generate_labels(
                 )
 
             labels[field_name] = scalar_label
+        elif (value is None and empty_labels) or (value is not None and value.location is None):
+            labels[field_name] = ScalarLabel(
+                Label=ScalarLabel.L(Source=[]),
+                Context=ScalarLabel.C(Entities=[]),
+                IsPrediction=False,
+                ModelVersion=model_versions.get(field_name, 0),
+            )
     return labels
 
 
@@ -292,10 +332,86 @@ def data_projection(skip_downloading_text):
         return "[uid, name: File.name, text: File.text, entities: File.ner.entities]"
 
 
+def escape_name(name):
+    return name.replace("'", "\\'")
+
+
+def upload_files(conn, collection_uid, b):
+    names = ", ".join([f"'{escape_name(f['name'])}'" for f in b])
+    files_by_name = {}
+    for f in b:
+        files_by_name[f["name"]] = f
+
+    all_uids = {}
+    existing = conn.query(f"@`file_collections::{collection_uid}`[uid, name: File.name] in(name, {names})")["data"]
+    for e in existing:
+        all_uids[e["name"]] = e["uid"]
+        del files_by_name[e["name"]]
+
+    fb_v = [x for x in files_by_name.items()]
+    upload_uids = conn.upload_files(collection_uid, [f for (_, f) in fb_v])
+    for (uid, (fname, _)) in zip(upload_uids, fb_v):
+        all_uids[fname] = uid
+
+    uids = set(upload_uids)
+    cursor = None
+    for i in range(10):
+        if len(uids) == 0:
+            return [all_uids[f["name"]] for f in b]
+
+        uid_filter = "in(uid, %s)" % (", ".join(['"%s"' % u for u in uids]))
+        resp = conn.query(
+            f"@`file_collections::{collection_uid}`[uid] {uid_filter} File.IsPreprocessed=true",
+            mode="poll",
+            timeout=360,
+            cursor=cursor,
+        )
+        cursor = resp["cursor"]
+
+        for d in resp["data"] or []:
+            if d["action"] != "insert":
+                continue
+
+            uid = d["data"]["uid"]
+            uids.remove(uid)
+    assert False, "Poll timed out after an hour"
+
+
+def retrieve_text(conn, collection_uid, uid_batch, skip_downloading_text):
+    results = {}
+
+    uids = set(uid_batch)
+    cursor = None
+    for i in range(10):
+        if len(uids) == 0:
+            return [results[uid] for uid in uid_batch]
+
+        uid_filter = "in(uid, %s)" % (", ".join(['"%s"' % u for u in uids]))
+        resp = conn.query(
+            f"@`file_collections::{collection_uid}`{data_projection(skip_downloading_text)} {uid_filter} File.IsPreprocessed=true",
+            mode="poll",
+            timeout=360,
+            cursor=cursor,
+        )
+        cursor = resp["cursor"]
+
+        for d in resp["data"] or []:
+            if d["action"] != "insert":
+                continue
+
+            uid = d["data"]["uid"]
+            results[uid] = d["data"]
+            uids.remove(uid)
+
+    assert False, "Poll timed out after an hour"
+
+
+# NOTE: Deprecated
 def upload_and_retrieve_text(conn, collection_uid, f, skip_downloading_text):
     existing = conn.query(
         "@`file_collections::%s`%s name='%s' File.IsPreprocessed=true"
         % (collection_uid, data_projection(skip_downloading_text), f["name"].replace("'", "\\'")),
+        timeout=360,
     )["data"]
     if len(existing) > 0:
         return existing[0]
@@ -308,7 +424,7 @@ def upload_and_retrieve_text(conn, collection_uid, f, skip_downloading_text):
                 "@`file_collections::%s`%s uid='%s' File.IsPreprocessed=true"
                 % (collection_uid, data_projection(skip_downloading_text), uids[0]),
                 mode="poll",
-                timeout=60,
+                timeout=360,
             )
 
             for d in resp["data"] or []:
@@ -335,9 +451,16 @@ def fields_to_doc_schema(fields) -> DocSchema:
     ret = {}
     for f in fields:
         comment = json.loads(f["comment"])
-        trainer = (
-            InferredFieldType.match_trainer(comment["infer_func"]["trainer_name"]) if "infer_func" in comment else None
-        )
+        try:
+            trainer = (
+                InferredFieldType.match_trainer(comment["infer_func"]["trainer_name"])
+                if "infer_func" in comment
+                else None
+            )
+        except AssertionError:
+            # If this is an unknown trainer, e.g. table v1, just skip the field
+            continue
+
         t = None
         if trainer == InferredFieldType.table:
             sub_fields = find_path(f, "Label", "Value", "Label", "Value").get("children", [])
@@ -371,64 +494,113 @@ def fields_to_doc_schema(fields) -> DocSchema:
 
 @validate_arguments
 def maybe_get_location(label: ScalarLabel, field_type: str) -> Optional[Location]:
-
     if label.Label.Source is None:
         return None
 
+    uids = []
+
     if field_type in ("CheckboxLabel", "SignatureLabel"):
         locations = label.Label.Source.BBoxes
     elif field_type in ("TextLabel", "NumberLabel", "TimestampLabel"):
         locations = [l.location for l in label.Label.Source]
+        uids = [l.uid for l in label.Label.Source]
     elif field_type == "DocumentTagLabel":
         locations = None
     else:
         raise ValueError(f"Unable to retrieve location from label of type {field_type} ({label = })")
 
     if locations:
-        return combine_locations(locations)
+        loc = combine_locations(locations)
+        loc.uids = uids
+        return loc
     return None
 
 
 @validate_arguments
 def maybe_get_value(label: ScalarLabel, field_type: str) -> Optional[Any]:
-
     value = label.Label.Value
     if field_type in ("CheckboxLabel", "SignatureLabel"):
         value = value["State"]
     elif field_type == "TimestampLabel" and value is not None:
         value = parse_date(value)
+    elif field_type == "DocumentTagLabel":
+        value = [e["Label"]["Value"] for e in value if e["Label"] and e["Label"]["Value"] is not None]
 
     return value
 
 
 @validate_arguments
-def row_to_record(row, doc_schema: DocSchema) -> Any:
+def row_to_record(
+    log,
+    row,
+    doc_schema: DocSchema,
+    allow_predictions: bool,
+    allow_low_confidence: bool,
+    reverse_field_mapping: Dict[str, str],
+) -> Any:
     d = {}
-    for field_name, field_type in doc_schema.fields.items():
+    for mapped_name, field_type in doc_schema.fields.items():
+        field_name = reverse_field_mapping.get(mapped_name, mapped_name)
+
         label = None
-        field = row.get(field_name, None)
+        field = row.get(field_name)
         if isinstance(field_type, DocSchema):
-            table_rows = [row_label["Label"]["Value"] for row_label in field["Label"]["Value"]]
-            label = [x for x in [row_to_record(tr, field_type) for tr in table_rows] if x is not None]
+            label = None
+            if field is not None and field.get("Label") and field.get("Label").get("Value"):
+                table_rows = [
+                    row_label["Label"]["Value"]
+                    for row_label in field["Label"]["Value"]
+                    if row_label and row_label.get("Label")
+                ]
+                label = [
+                    x
+                    for x in [
+                        row_to_record(
+                            log, tr, field_type, allow_predictions, allow_low_confidence, reverse_field_mapping
+                        )
+                        for tr in table_rows
+                    ]
+                    if x is not None
+                ]
             if not label:
                 continue
-        elif field is not None:
-            impira_label = ScalarLabel.parse_obj(field)
-            if impira_label.Label.IsPrediction:
+        elif field is not None and field.get("Label") is not None:
+            try:
+                impira_label = ScalarLabel.parse_obj(field)
+            except ValidationError as e:
+                log.warning(
+                    f"Record with uid={row['uid']} has an invalid label for field `{field_name}`: {e}. Skipping..."
+                )
+                continue
+
+            if impira_label.Label.IsPrediction and not (
+                allow_predictions and (allow_low_confidence or impira_label.Label.IsConfident)
+            ):
                 continue
 
             location = maybe_get_location(impira_label, field_type)
-            value = maybe_get_value(impira_label, field_type)
+
+            try:
+                value = maybe_get_value(impira_label, field_type)
+            except Exception as e:
+                log.warning(
+                    f"Record with uid={row['uid']} has an invalid label for field `{field_name}`. Failed to parse: {e}. Skipping..."
+                )
+                continue
 
             label = {"location": location, "value": value}
 
-        d[field_name] = label
+        d[mapped_name] = label
 
     if len(d) == 0:
         return None
 
     M = schema_to_model(doc_schema)
-    return M(**d)
+    try:
+        return M(**d)
+    except ValidationError as e:
+        log.warning(f"Record with uid={row['uid']} has an invalid label: {e}")
+        return None
 
 
 @validate_arguments
@@ -447,6 +619,13 @@ def row_to_fname(row, use_original_filename) -> str:
         return fname.replace("/", "_")
 
 
+def fname_filter(fnames):
+    return "in(File.name, %s)" % (",".join(['"%s"' % n.replace('"', '\\"') for n in fnames]))
+
+
+RETRIES = 10
+
+
 class Impira(Tool):
     class Config(BaseModel):
         api_token: str
@@ -483,12 +662,20 @@ class Impira(Tool):
         add_files=False,
         skip_missing_files=False,
         skip_new_fields=False,
+        empty_labels=False,
         collection_name=None,
         max_fields=-1,
+        first_file=0,
         max_files=-1,
+        batch_size=50,
+        first_batch=0,
+        cache_dir: pathlib.Path = None,
     ):
         log = self._log()
 
+        if cache_dir:
+            cache_dir.mkdir(parents=True, exist_ok=True)
+
         schema = generate_schema(doc_schema)
 
         skip_downloading_text = all([t == "DocumentTagLabel" for t in doc_schema.fields.values()])
@@ -514,39 +701,87 @@ class Impira(Tool):
 
         assert (not add_files) or skip_upload, "Cannot add existing files to the collection unless you skip upload"
 
+        new_entries = [e for e in entries]
+        new_entries.sort(key=lambda e: e.record is None)  # Place the rows with records up front
+        entries = new_entries[first_file:]
         if max_files != -1:
-            new_entries = [e for e in entries]
-            new_entries.sort(key=lambda e: e.record is None)  # Place the rows with records up front
             entries = new_entries[:max_files]
 
         if not skip_upload:
             files = [{"path": e.url or str(e.fname), "name": e.fname.name} for e in entries]
 
             log.info("Uploading %d files", len(files))
+            with ThreadPoolExecutor(max_workers=parallelism) as t:
+                uid_batches = [
+                    x for x in t.map(lambda f: upload_files(self._conn(), collection_uid, f), batch(files, 20))
+                ]
+
+            log.info("Retrieving text for %d files", len(files))
             with ThreadPoolExecutor(max_workers=parallelism) as t:
                 file_data = [
                     x
-                    for x in t.map(
-                        lambda f: upload_and_retrieve_text(self._conn(), collection_uid, f, skip_downloading_text),
-                        files,
+                    for results_batch in t.map(
+                        lambda b: retrieve_text(self._conn(), collection_uid, b, skip_downloading_text),
+                        uid_batches,
                     )
+                    for x in results_batch
                 ]
+
+            if cache_dir:
+                for f, record in zip(files, file_data):
+                    cache_file = cache_dir / f"{f['name']}.json"
+                    with open(cache_file, "w") as f:
+                        json.dump(record, f)
         else:
+            all_fnames = [e.fname.name for e in entries]
+            cached = 0
+            missing_file_uids = {}
+            log.info("Retrieving text for %d files", len(all_fnames))
             while True:
-                uids = {
-                    r["name"]: r
-                    for r in conn.query(
-                        "@`file_collections::%s`%s" % (collection_uid, data_projection(skip_downloading_text))
-                    )["data"]
-                }
+                uids = {}
+                for b in batch(all_fnames, 50):
+                    remaining = []
+
+                    if cache_dir:
+                        for name in b:
+                            cache_file = cache_dir / f"{name}.json"
+                            if name not in missing_file_uids and cache_file.exists():
+                                with open(cache_file, "r") as f:
+                                    record = json.load(f)
+                                    if record is not None:
+                                        uids[name] = record
+                                cached += 1
+                            else:
+                                remaining.append(name)
+
+                    new_records = (
+                        {
+                            r["name"]: r
+                            for r in conn.query(
+                                "@`file_collections::%s`%s %s"
+                                % (collection_uid, data_projection(skip_downloading_text), fname_filter(remaining))
+                            )["data"]
+                        }
+                        if remaining
+                        else {}
+                    )
+
+                    uids.update(new_records)
+                    if cache_dir:
+                        for fname in remaining:
+                            cache_file = cache_dir / f"{fname}.json"
+                            if fname in new_records:
+                                with open(cache_file, "w") as f:
+                                    json.dump(new_records[fname], f)
+                            else:
+                                with open(cache_file, "w") as f:
+                                    json.dump(None, f)
 
                 if add_files:
                     missing_files = [e.fname.name for e in entries if e.fname.name not in uids]
                     if len(missing_files) == 0:
                         break
-                    file_filter = "in(File.name, %s)" % (
-                        ",".join(['"%s"' % n.replace('"', '\\"') for n in missing_files])
-                    )
+                    file_filter = fname_filter(missing_files)
                     missing_file_uids = {
                         r["name"]: r["uid"]
                         for r in conn.query("@files[name: File.name, uid] %s" % (file_filter))["data"]
@@ -565,6 +800,7 @@ class Impira(Tool):
                 else:
                     break
 
+            log.info("Done retrieving text for %d files (%d cached, %d final)", len(all_fnames), cached, len(uids))
             if skip_missing_files:
                 entries = [e for e in entries if e.fname.name in uids]
 
@@ -575,10 +811,12 @@ class Impira(Tool):
         # Now, just trim it down to the labeled entries
         labeled_entries = []
         labeled_files = []
+        seen_uids = set()
         for i, e in enumerate(entries):
-            if e.record is not None:
+            if e.record is not None and file_data[i]["uid"] not in seen_uids:
                 labeled_entries.append(e)
                 labeled_files.append(file_data[i])
+                seen_uids.add(file_data[i]["uid"])
 
         if len(labeled_entries) == 0:
             log.warning("No records have labels. Stopping now that uploads have completed.")
@@ -597,18 +835,26 @@ class Impira(Tool):
         }
 
         labels = []
-        for e, fd in zip(labeled_entries, labeled_files):
-            entity_map = EntityMap(entities=fd["entities"])
-            labels.append(
-                generate_labels(
-                    log,
-                    fd["name"],
-                    e.record,
-                    fd["text"]["words"],
-                    entity_map,
-                    model_versions,
+        for i, (e, fd) in enumerate(zip(labeled_entries, labeled_files)):
+            try:
+                words = fd["text"]["words"]
+                word_map = {w["uid"]: w for w in words}
+                entity_map = EntityMap(entities=fd["entities"] or [])
+                labels.append(
+                    generate_labels(
+                        log,
+                        fd["name"],
+                        e.record,
+                        fd["text"]["words"],
+                        word_map,
+                        entity_map,
+                        model_versions,
+                        empty_labels,
+                    )
                 )
-            )
+            except Exception as e:
+                log.warning(f"Unable to process record (uid={fd['uid']}): %s", e)
+                labels.append({})
 
         schema_resp = conn.query("@file_collections::%s limit:0" % (collection_uid))
         current_fields = fields_to_doc_schema(filter_inferred_fields(schema_resp["schema"]["children"])).fields
@@ -633,16 +879,16 @@ class Impira(Tool):
                             ]
                         )
 
-                    if InferredFieldType.timestamp in unique_entity_types:
-                        narrow_type = InferredFieldType.timestamp
-                    elif InferredFieldType.number in unique_entity_types:
-                        narrow_type = InferredFieldType.number
-                    else:
-                        narrow_type = field_type
+                        if InferredFieldType.timestamp in unique_entity_types:
+                            narrow_type = InferredFieldType.timestamp
+                        elif InferredFieldType.number in unique_entity_types:
+                            narrow_type = InferredFieldType.number
+                        else:
+                            narrow_type = field_type
 
-                    narrow_types.add(narrow_type)
-                    if narrow_type == InferredFieldType.text:
-                        break
+                        narrow_types.add(narrow_type)
+                        if narrow_type == InferredFieldType.text:
+                            break
 
                 for weak_type in [InferredFieldType.text, InferredFieldType.number, InferredFieldType.timestamp]:
                     if weak_type in narrow_types:
@@ -656,7 +902,7 @@ class Impira(Tool):
                     continue
 
                 existing_label_type = label_name_to_inferred_field_type(existing_field)
-                if existing_label_type != field_type:
+                if existing_label_type != field_type and not skip_type_inference:
                     log.warning(
                         "Field %s already created with type %s, but we're setting it to a value of type %s. You may want to delete it.",
                         f.name,
@@ -680,61 +926,31 @@ class Impira(Tool):
         log.info("Creating fields: %s" % (field_names_to_update))
 
         if len(field_specs) > 0:
-            conn.create_fields(collection_uid, field_specs)
+            for b in batch(field_specs, n=5):
+                conn.create_fields(collection_uid, b)
 
         fields_to_update = [f for f in schema if len(f.path) == 0 and f.name in field_names_to_update]
 
-        # Snapshot the model version and its cursor for new fields. Old fields are not going to change
-        # their values (unless the labels have changed, which is not something we distinguish).
-        expected_increment_query = "@`file_collections::%s`[increment: %s]" % (
-            collection_uid,
-            " + ".join(
-                ["0"]
-                + [
-                    "(SUM(IF(`%s`.Label.IsPrediction or `%s`=null, 1, %d-`%s`.ModelVersion)))"
-                    % (
-                        f.name,
-                        f.name,
-                        model_versions.get(f.name, 0),
-                        f.name,
-                    )
-                    for (f) in fields_to_update
-                ]
-            ),
-        )
-
-        log.info("Running expected increment query %s", expected_increment_query)
-        increment = conn.query(expected_increment_query)["data"][0]["increment"]
-
-        mv_query = "@`file_collections::%s`[sum_mv: %s]" % (
-            collection_uid,
-            " + ".join(["0"] + ["SUM(`%s`.`ModelVersion`)" % (f.name) for f in fields_to_update]),
-        )
-
-        # Unfortunately, polling doesn't work for "min" queries, so we just run a normal query
-        log.info("Running model version query %s", mv_query)
-        resp = conn.query(mv_query)["data"][0]
-        target = resp["sum_mv"] + increment
-
-        log.info(
-            "Current minimum model version total %s. Target is %s across %d fields",
-            resp["sum_mv"],
-            target,
-            len(fields_to_update),
-        )
-
         log.info("Running update on %d files" % len(labeled_files))
 
-        # Batch the updates into chunks of 200, and retry a few times because of deadlock-issues
-        for b in batch([x for x in zip(labeled_files, labels)], n=200):
-            for i in range(10):
+        # Batch the updates into chunks and retry a few times because of deadlock-issues
+        batches = [b for b in batch([x for x in zip(labeled_files, labels)], n=batch_size)]
+        for b_idx, b in enumerate(batches):
+            if b_idx < first_batch:
+                continue
+            log.info("Updating batch %d/%d", b_idx, len(batches) - 1)
+            processed = 0
+            for i in range(RETRIES):
                 if i > 0:
                     log.warning("Sleeping for 1 second...")
                     time.sleep(1)
                 try:
-                    conn.update(
-                        collection_uid,
-                        [
+                    mini_batches = [x for x in batch(b[processed:], n=max(1, batch_size // (i + 1)))]
+
+                    for mb_idx, mb in enumerate(mini_batches):
+                        if len(mini_batches) > 1:
+                            log.info("Doing a mini-update (attempt %d): %d/%d", i, mb_idx, len(mini_batches) - 1)
+                        update_record = [
                             {
                                 **{"uid": fd["uid"]},
                                 **{
@@ -743,55 +959,61 @@ class Impira(Tool):
                                     if field_path in field_names_to_update
                                 },
                             }
-                            for (fd, ld) in b
-                        ],
-                    )
+                            for (fd, ld) in mb
+                        ]
+                        conn.update(collection_uid, update_record)
+                        processed += len(mb)
+                    if i > 0:
+                        log.info("Success!")
                     break
                 except APIError as e:
-                    log.warning("Failed to update. Will retry up to %d more times: %s" % (10 - i - 1, e))
+                    if i < RETRIES - 1:
+                        log.warning("Failed to update. Will retry up to %d more times: %s" % (10 - i - 1, e))
+                    else:
+                        raise
 
         log.info("Done running update on %d files. Models will now update!" % len(labeled_files))
 
-        # This code is currently too brittle to be useful, since model versions aren't a reliable way of knowing
-        # when evaluation has finished. We can reintroduce it once we have a better strategy in place.
-        #
-        #        while True:
-        #            resp = conn.query(
-        #                mv_query + " [sum_mv] sum_mv >= %d" % (target),
-        #            )
-        #            try:
-        #                resp = resp["data"][0]
-        #                break
-        #            except Exception as e:
-        #                time.sleep(1)
-        #                continue
-        #
-        #        # TODO: We know this undercounts a bit, so we should probably check the "spinner" query too
-        #
-        #        log.debug(
-        #            "Current minimum model version %s",
-        #            resp["sum_mv"],
-        #        )
-        log.info("Done!")
-
     @validate_arguments
     def snapshot(
-        self, collection_uid: str, use_original_filenames=False, labeled_files_only=False, filter_collection_uid=None
+        self,
+        collection_uid: str,
+        use_original_filenames=False,
+        labeled_files_only=False,
+        filter_collection_uid=None,
+        label_filter=None,
+        allow_low_confidence=False,
+        field_mapping: Optional[Dict[str, str]] = {},
     ):
         log = self._log()
 
         conn = self._conn()
 
-        filt = f"-join_one(`file_collections::{filter_collection_uid}`, uid, uid)=null" if filter_collection_uid else ""
+        collection_filter = (
+            f"-join_one(`file_collections::{filter_collection_uid}`, uid, uid)=null" if filter_collection_uid else ""
+        )
 
-        resp = conn.query("@`file_collections::%s` %s" % (collection_uid, filt))
+        allow_predictions = f"{label_filter}" if label_filter else "false"
+
+        resp = conn.query(
+            "@`file_collections::%s`[.*, __allow_predictions: %s] %s"
+            % (collection_uid, allow_predictions, collection_filter)
+        )
 
         doc_schema = fields_to_doc_schema(filter_inferred_fields(resp["schema"]["children"]))
+
+        if field_mapping is not None:
+            doc_schema.fields = {field_mapping.get(n): t for (n, t) in doc_schema.fields.items() if n in field_mapping}
+        else:
+            field_mapping = {}
+        reverse_field_mapping = {v: k for (k, v) in field_mapping.items()}
         records = [
             {
                 "url": row["File"]["download_url"],
                 "name": row_to_fname(row, use_original_filenames),
-                "record": row_to_record(row, doc_schema),
+                "record": row_to_record(
+                    log, row, doc_schema, bool(row["__allow_predictions"]), allow_low_confidence, reverse_field_mapping
+                ),
             }
             for row in resp["data"]
         ]
@@ -804,21 +1026,29 @@ class Impira(Tool):
         return doc_schema, records
 
     @validate_arguments
-    def snapshot_collections(self, use_original_filenames=False, max_files_per_collection=-1, num_samples=2):
+    def snapshot_collections(
+        self, use_original_filenames=False, max_files_per_collection=-1, num_samples=2, collection_filter=None
+    ):
         log = self._log()
 
+        full_collection_filter = "-collection=null"
+        if collection_filter:
+            full_collection_filter += " in(collection_uid, %s)" % (
+                ", ".join(["'%s'" % (uid) for uid in collection_filter])
+            )
+
         conn = self._conn()
         files = conn.query("@files[uid, File: File[download_url, name]] -File.download_url=null -`File type`=Data")[
             "data"
         ]
         collections = conn.query(
-            "@file_collection_contents[collection_uid, files: array_agg(file_uid)] -collection=null"
+            f"@file_collection_contents[collection_uid, files: array_agg(file_uid)] {full_collection_filter}"
         )["data"]
 
         collection_names = {}
-        for row in conn.query("@file_collection_contents[collection_uid, name: collection.name] -collection=null")[
-            "data"
-        ]:
+        for row in conn.query(
+            f"@file_collection_contents[collection_uid, name: collection.name] {full_collection_filter}"
+        )["data"]:
             collection_names[row["collection_uid"]] = row["name"]
 
         file_membership = {}
@@ -856,12 +1086,12 @@ class Impira(Tool):
                 "url": row["File"]["download_url"],
                 "name": row_to_fname(row, use_original_filenames),
                 "record": {
-                    "Doc tag": DocumentTagLabel(value=collection_names[file_membership[row["uid"]][0]])
+                    "Doc tag": DocumentTagLabel(value=[collection_names[x] for x in file_membership[row["uid"]]])
                     if row["uid"] in file_membership and len(file_membership[row["uid"]]) == 1
-                    else DocumentTagLabel(value=None)
+                    else DocumentTagLabel(value=[])
                     if row["uid"] not in file_membership
                     else None,
-                    "Sampled tag": DocumentTagLabel(value=collection_names[sampled[row["uid"]]])
+                    "Sampled tag": DocumentTagLabel(value=[collection_names[sampled[row["uid"]]]])
                     if row["uid"] in sampled
                     else None,
                 },
